/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "uip-appcall.h"
#include "avrutils.h"
#include "uip/uip.h"
#include "amish.h"
#include "amish_uip.h"
#include <string.h>

/* Authentication password: */
char amish_uip_pw[AMISH_PW_LEN+1];

void amish_uip_init(void) {
	/* Create TCP telnet listening port for the AMISH */
	uip_listen(HTONS(AMISH_TCP_PORT));
	DEBUG_PUT(2, "AMISH bound.");
}

#define amish_uip_send_pw_prompt() \
	{strcpy_P(s->txbuf, PSTR("Password: ")); s->txlen = 10;\
	 }

void amish_uip_send_prompt(void) {
	struct amish_uip_state *s = (struct amish_uip_state *)&(uip_conn->appstate);
	s->txlen = stpcpy_P(s->txbuf, PSTR(AMISH_PROMPT "> ")) - s->txbuf;
	s->state = AMISH_UIP_STATE_PROMPT_SENT;
}

void amish_uip_check_password(void) {
	struct amish_uip_state *s = (struct amish_uip_state *)&(uip_conn->appstate);
	if(!strcmp(s->rxbuf, amish_uip_pw)) {
		/* Empty buffer: */
		s->rxlen = 0;
		amish_uip_send_prompt();
	} else
		uip_abort();
}

/* This function is called if we received data */
void amish_uip_receive(void) {
	struct amish_uip_state *s = (struct amish_uip_state *)&(uip_conn->appstate);

	if(s->rxbuf[0] == '\0' && s->rxlen <= 2) {
		/* empty string entered */
		s->rxlen = 0;
		amish_uip_send_prompt();
	} else if(s->state != AMISH_UIP_STATE_CMD_PROCESSING) {
		/* Magic command to end the connection? */
		if(!strcmp_P(s->rxbuf, PSTR("exit"))) {
			uip_abort();
			return;
		}

		/* No, we received a different command. Mark the state and process it.*/
		amish_cmd_arg ca;
		ca.input = s->rxbuf;
		ca.output = s->txbuf;
		ca.len = AMISH_UIP_TXBUF;
		s->cmd = amish_firstexec_cmd(&ca);

		amish_io_len retlen = ca.ret.len;
		/* Append a linefeed for user convenience. */
		s->txbuf[retlen++] = '\r';
		s->txbuf[retlen++] = '\n';
		s->txlen = retlen;
		s->cmd_state = ca.state;
		s->reenter_cmd = ca.ret.call_again;
		s->state = AMISH_UIP_STATE_CMD_PROCESSING;
	}

}

/* This function is called if an ack for previous data was received */
void amish_uip_ack(void) {
	struct amish_uip_state *s = (struct amish_uip_state *)&(uip_conn->appstate);

	if(!s->reenter_cmd) {
		/* Data successfully transfered, mark the buffer as free. */
		s->rxlen = 0;
		s->txlen = 0;
		amish_uip_send_prompt();
	} else {
		/* Generate the next chunk */
		amish_cmd_arg ca;
		ca.input = s->rxbuf;
		ca.output = s->txbuf;
		ca.len = AMISH_UIP_TXBUF - s->txlen;
		ca.state = s->cmd_state;
		amish_reexec_cmd(s->cmd, &ca);
		amish_io_len retlen = ca.ret.len;
		s->txbuf[retlen++] = '\r';
		s->txbuf[retlen++] = '\n';
		s->txlen = retlen;
		s->reenter_cmd = ca.ret.call_again;
		s->cmd_state = ca.state;
	}
}

void amish_uip_appcall(void) {
	struct amish_uip_state *s = (struct amish_uip_state *)&(uip_conn->appstate);

	/* Whoo, a new connection for us! */
	if(uip_connected()) {
		/* Blank the state of this connection. */
		DEBUG_PUT(1, "AMISH connected.");
		s->rxlen = 0;
		s->txlen = 0;
		s->reenter_cmd = 0;
		if(amish_uip_pw[0] == '-' && !amish_uip_pw[1]) {
			amish_uip_send_prompt();
		} else {
			s->state = AMISH_UIP_STATE_CONNECTED;
			amish_uip_send_pw_prompt();
		}
	}

	if(uip_acked()) {
		/* In every case, our packet is transmittet, so forget it: */
		s->txlen = 0;

		switch(s->state) {
			case AMISH_UIP_STATE_CONNECTED:
				break;
			case AMISH_UIP_STATE_PW_SENT:
				amish_uip_send_prompt();
				s->state = AMISH_UIP_STATE_PROMPT_SENT;
				break;
			case AMISH_UIP_STATE_PROMPT_SENT:
				break;
			case AMISH_UIP_STATE_CMD_PROCESSING:
				amish_uip_ack();
				break;
		}
	}

	if(uip_newdata()) {
		uint16_t pkt_len = uip_datalen();
		char* linefeed;

		if(s->state == AMISH_UIP_STATE_CONNECTED && s->rxlen == 0
			&& ((char*)uip_appdata)[0] > 250)
			/* Discard. This is a telnet configuration packet. */
			return;


		amish_io_len rxlen = s->rxlen;
		if(pkt_len + rxlen < AMISH_UIP_RXBUF) {
			/* There is enough space in the buffer, copy the bytes over */
			memcpy(s->rxbuf + rxlen, uip_appdata, pkt_len);

			rxlen += pkt_len;
			s->rxlen = rxlen;

			/* Check for a line ending */
			if((linefeed = (char*) memchr(s->rxbuf, '\n', rxlen))) {
				*linefeed = '\0'; /* Commands end on '\0' internally, not on '\n'/'\r' */

				/* Found one. Clear an eventual \r occurence and process the command. */
				if((linefeed = (char*) memchr(s->rxbuf, '\r', rxlen)))
					*linefeed = '\0';

				/* The line is complete. What now? */
				switch(s->state) {
					case AMISH_UIP_STATE_CONNECTED:
						amish_uip_check_password();
						break;
					case AMISH_UIP_STATE_PROMPT_SENT:
					case AMISH_UIP_STATE_CMD_PROCESSING:
						amish_uip_receive();
						break;
				}
			}
		} else {
			/* Too much data received, close the connection */
			uip_abort();
		}
	}

	/* if(closed, timeout, abort): ignore, nothing to do since we are stateless */

	/*
	 * In the case of the previous events or polling or retransmitting,
	 * check if we have a packet to send.
 	 */
	if(uip_connected() | uip_acked() | uip_newdata() | uip_rexmit() | uip_poll()){
		/* Is the TX buffer filled? */
		if(s->txlen) {
			uip_send(s->txbuf, s->txlen);
		}
	}
}
