/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>
#include "config.h"
#include "dcf77.h"
#include "timer.h"
#include "avrutils.h"


/* The minimal length of a signal segment to be counted as DCF77 impulse (in ms): */
#define DCF77_MIN_PULSE_LENGTH 900
#define DCF77_MAX_PULSE_LENGTH 1020 /* Pulse length for minute end condition */
#define DCF77_MIN_LAST_PULSE_LENGTH 1900 /* Minute end impulse length */
#define DCF77_MIN_DOWN_LENGTH 25 /* ms; so long must a low impulse be to be counted */

/*
 * DCF77 averaging interval for offset correction in seconds:
 * A minute has only 59 falling edges, not 60.
 */
#define DCF77_AVR_INTERVAL 5*59

/* Delay added to reception delay when synchronizing the local clock: */
#define DCF77_SET_DELAY 492 /* microseconds */

/* There are only two datagrams for now: The current and the cumulative datagram. */
#define DCF77_DATAGRAMS 2
/* Standard datagram where incoming bits are put: */
#define DCF77_STD_DGRAM 0
/* Data will be collected for error correction in this datagram: */
#define DCF77_CUMULATIVE_DGRAM 1

#define EDGE_DOWN 0
#define EDGE_UP 1

/*
 * 2^(k+5) is 16 times more than 2^k.
 * So, subtract the upper nibble 6 times to count it 10 times instead
 * of 16 times. This converts BCD to binary:
 */
static noinline uint8_t dcf77_bcd2bin(uint8_t bcd){
	uint8_t upper_nibble = bcd >> 4;
	/* return bcd - 6*bcd */
	return (bcd - (upper_nibble <<2) - (upper_nibble <<1));
}

/*
 * Sequence of masks for the first n bits in a byte:
 * A full byte mask is not needed.
 */
static const uint8_t dcf77_byte_mask[8] PROGMEM = {0, 1, 3 ,7, 15, 31, 63, 127};

/* Length of a datagram in bytes: */
#define DGRAM_LEN 8
/* Memory for the DCF77 datagrams and validation state: */
uint8_t dcf77_datagram[DCF77_DATAGRAMS][DGRAM_LEN];
uint8_t dcf77_datagram_valid[DCF77_DATAGRAMS][DGRAM_LEN];

/* State of the DCF77 decoder: */
struct dcf77_info dcf77_info;

/* DCF77 date/time derived from datagram */
date_time timer_dcf77;

/* DCF77 signal statistics: */
uint32_t mean_edge_ticks; /* Sum of the ticks at the falling edge of one minute */
uint16_t mean_edge_counter; /* Number of values in mean_edge_ticks */

uint16_t dcf77_recv_delay; /* Total reception delay in ticks */

/* Structure to return DCF77 bitranges: */
struct dcf77_byte {
	uint8_t data; /* Represents the wanted bits */
	uint8_t valid:2; /* 0: invalid, 1: valid, 2: valid bits from cumulative datagram */
};

/* Timestamp at the last level change at the signal input: */
timer_millistamp dcf77_t_edge;
/* Total length of the current impulse and length of its low-part (in ms): */
uint16_t current_pulse_length, current_down_length;

static void dcf77_process_second(void);
static struct dcf77_byte dcf77_get_bits(uint8_t start, uint8_t nr_bits);
void dcf77_decode_datagram(void);
static uint8_t dcf77_check_parity(uint32_t value, uint8_t parbitno);
static uint8_t dcf77_bcd2bin(uint8_t bcd);
static void dcf77_set_dgram_bit(uint8_t dgram[][DGRAM_LEN], uint8_t dgram_no,
	uint8_t dgram_bit, uint8_t value);
static void dcf77_process_stats(void);
static uint8_t dcf77_pulse_ok(uint16_t level_length);
static void dcf77_clear_edge_stats(void);
static uint8_t dcf77_seconds(void);

/*
 * dcf77_init:
 * Initialize the DCF77 subsystem.
 */
void dcf77_init(void) {
#if USE_PULSE_ICP
	/* Configure ICP1:
	 * - enable noise canceler (ICNC1 = 1)
	 * - trigger on falling edges (ICES1 = 0)
	 * - no interrupts for a trigger (ICIE1 = 0)
	 */
	TCCR1B &= ~((1 <<	ICNC1) | (1 <<	ICES1));
	TCCR1B |= 1 << ICNC1;
	TIMSK1 &= ~(1 <<	ICIE1);
#endif

	bzero(&dcf77_info, sizeof(dcf77_info));
	bzero(&dcf77_t_edge, sizeof(timer_millistamp));
	bzero(dcf77_datagram, 8 * DCF77_DATAGRAMS);
	bzero(dcf77_datagram_valid, 8 * DCF77_DATAGRAMS);
	current_pulse_length = current_down_length = 0;
	/* We start on a rising edge: */
	dcf77_info.edge = EDGE_UP;

	/* The sampling begins at second 1, not 0, to enforce an invalid first datagram: */
	dcf77_info.second = 1;
	dcf77_clear_edge_stats();
	DEBUG_PRINT(1, "DCF77 reception delay: %u ticks\n", dcf77_recv_delay);
}

/*
 * dcf77_worker:
 * Handle signal level transitions and reconstruct DCF77 data frames.
 */
void dcf77_worker(void) {
	uint8_t edge = dcf77_info.edge;
	if(IS_SET(DCF77_PULSE_PIN) == edge) {
		timer_millistamp t_temp;
		timer_get_timestamp(&t_temp);
		uint16_t tick;
#if USE_PULSE_ICP
		tick = ICR1;
#else
		tick = t_temp.ms;
#endif

		DEBUG_PRINT(3, "I: %u E%u\n", IS_SET(DCF77_PULSE_PIN), edge);
		/*
		 * The last frame was valid. It is time to sync and set the time
		 * after processing the frame.
		 */
		uint8_t second_top = dcf77_seconds();
		if(edge == EDGE_DOWN && !dcf77_info.timer_in_sync &&
		 dcf77_info.second >= second_top && dcf77_info.time_valid) {
			timer_sync_ticks(dcf77_recv_delay + timer_ctr_us2ticks(DCF77_SET_DELAY));
			DEBUG_PRINT(1, "DCF77: Timer synchronised (Offset %u).\n", dcf77_recv_delay + DCF77_SET_DELAY);
			dcf77_info.set_time = 1;
		} else
			dcf77_info.set_time = 0;

		uint16_t level_length = timer_diff_timestamp16(&dcf77_t_edge, &t_temp);
		/* Regardless of the signal level, count it as pulse length. */
		current_pulse_length += level_length;
		if(edge == EDGE_UP) { /* Signal edge down -> up. */
			/* Count the down level as down length */
			current_down_length += level_length;
		} else { /* edge up -> down */
			if(dcf77_pulse_ok(level_length)) {
				/*
				 * We have not captures one second of the signal, but a falling edge
				 * occured. This could be because of heavy signal interferences.
				 * Try to accumulate another rising and falling edge.
				 */
				DEBUG_PRINT(1, "DCF77: %lu Ignore %u ms, total pulse len: %u ms\n", timer_unix_timestamp, level_length, current_pulse_length);
				/*
				 * Heuristic to handle short signal spikes:
				 * If there was only a spike of less than 25 ms length detected,
				 * count it as if the signal has been low during this time.
				 * This assumes that such short spikes are transients from flipping
				 * a switch etc, and not part of the signal.
				 */
				if(level_length < DCF77_MIN_DOWN_LENGTH)
					current_down_length += level_length;
			} else { /* The pulse was long enough. */
				/* Process the impulse and clear the counters for the next one. */

				/*
				 * Fake the counter value if the time was set (== counter value set
				 * to dcf77_recv_delay + timer_ctr_us2ticks(DCF77_SET_DELAY)).
				 * Without this, the first pulse after setting the time would be
				 * too short, leading to a faulty datagram.
				 */
				if(dcf77_info.set_time)
					t_temp.ms = dcf77_recv_delay + timer_ctr_us2ticks(DCF77_SET_DELAY);

				/* Signal statistics: count this down pulse. */
				/* Handle a wrap-around by calculating the correct difference: */
				if(tick > TIMER_SCALE_VALUE/2)
					tick = TIMER_SCALE_VALUE - tick;
				add32_16(&mean_edge_ticks, tick);
				mean_edge_counter++;

				dcf77_process_second();

				dcf77_process_stats();

				current_pulse_length = 0;
				current_down_length = 0;
			}
		}
		/* Update the level change timestamp: */
		dcf77_t_edge = t_temp;
		dcf77_info.edge = !edge;
	}
}

/*
 * dcf77_process_second:
 * Process the gathered information.
 */
static void dcf77_process_second(void){
	uint8_t valid, time_valid;
	uint8_t long_pulse = 0, minute_end = 0;

	/*
	 * A pulse is only valid if its length is within strict bounds for a normal
	 * impulse or if it is very long (meaning: end-of-minute marker)
	 * Also, the low-time of the impulse must lay between the values of a
	 * logical 0 and 1.
	 */
	valid = (current_pulse_length > 900 && current_pulse_length < 1100);
	if(current_pulse_length > 1500) valid = 1;
	if(!(current_down_length > 80 && current_down_length < 220))
		valid = 0;


	/*
	 * A long pulse is counted as end-of-minute, 91 to 149 ms as logical 0,
	 * 150 to 209 as logical 1.
	 */
	if(current_pulse_length > 1500)
		minute_end = 1;
	if(current_down_length < 150)
		long_pulse = 0;
	else
		long_pulse = 1;

	uint8_t second = dcf77_info.second;
	DEBUG_PRINT(2, "DCF77 Impulse %u. down %u of %u ms (%svalid): l%hu, me%hu\n",
		second, current_down_length, current_pulse_length,
		valid?"":"in", long_pulse, minute_end);
	/* Write current bit into the datagram: */
	dcf77_set_dgram_bit(dcf77_datagram, 0, second, long_pulse);
	dcf77_set_dgram_bit(dcf77_datagram_valid, 0, second, valid);
	/* Merge the current datagram with previous ones to correct missing bits: */
	if(valid) {
		dcf77_set_dgram_bit(dcf77_datagram, DCF77_CUMULATIVE_DGRAM,
			second, long_pulse);
		dcf77_set_dgram_bit(dcf77_datagram_valid, DCF77_CUMULATIVE_DGRAM,
			second, valid);
	}

	second++;
	dcf77_info.second = second;
	dcf77_decode_datagram();

	/*
	 * With the end-of-minute mark, check for datagram completeness and validity.
	 */
	uint8_t second_top = dcf77_seconds();
	uint8_t dgram_end = second > second_top;
	if(minute_end && dgram_end) {
#if DEBUG_LEVEL > 2
		DEBUG_PRINT(3, "Datagram assembled: ");
		for(uint8_t i = 0; i < 60 ; i++){
			if(!(dcf77_datagram_valid[0][i/8] & (1 << (i % 8))))
				putchar('X');
			else if(dcf77_datagram[0][i/8] & (1 << (i % 8)))
				putchar('1');
			else
				putchar('0');
			if(!((i+1)%8)) DEBUG_PRINT(2, " ");
		}
		DEBUG_PRINT(3, "\n");
		DEBUG_PRINT(3, "Cumulative assembled: ");
		for(uint8_t i = 0; i < 60 ; i++){
			if(!(dcf77_datagram_valid[1][i/8] & (1 << (i % 8))))
				putchar('X');
			else if(dcf77_datagram[1][i/8] & (1 << (i % 8)))
				putchar('1');
			else
				putchar('0');
			if(!((i+1)%8)) DEBUG_PRINT(2, " ");
		}
		DEBUG_PRINT(3, "\n");
#endif

		if(dcf77_info.time_valid) {
			/* Should we set the local time? */
			if(dcf77_info.set_time) {
				dcf77_clear_sync();
				timer_set_time(&timer_dcf77);
				dcf77_info.timer_in_sync = 1;
				dcf77_info.first_set = 1;
				DEBUG_PRINT(0, "DCF77: Time set.\n");
				time_valid = 1;

			} else if(dcf77_info.timer_in_sync) { /* Check the received time against the local */
				uint8_t decis;
				time_t local_unix_time;

				/* Retrieve the local time: */
				ENTER_CRITICAL_SECTION
					decis = timer_info.decisecond_counter;
					local_unix_time = timer_unix_timestamp;
				LEAVE_CRITICAL_SECTION

				/*
				 * The DCF77 receiver and the low pass at the AVR's DCF77 level input
				 * delay the received signal. This ensures that the timer ISR finished
				 * at this point and the received time should equal the local one.
				 */

				if(decis != 0) {
					/* Hmm. Either, the timer ISR didn't run or the local clock is off
					 * by more than one decisecond. The first case does not happen
					 * because of the reception delay. Clear the validation state.
					 */
					DEBUG_PRINT(1, "DCF77: Local time (.%hhu) async.\n", decis);
					time_valid = 0;
					dcf77_info.invalid_frames = DCF77_MAX_INVALID_FRAMES;
				} else { /* decis == 0, as it should be. */
					time_t dcf77_unix_time = timer_localtime2unix(&timer_dcf77);
					if(dcf77_unix_time != local_unix_time) {
						DEBUG_PRINT(1, "DCF77: Local time != recv.\n");
						DEBUG_PRINT(2, "DCF77: Local time %lu recv %lu\n", local_unix_time, dcf77_unix_time);
						time_valid = 0;
					} else {
						/* The local time equals the received time. */
						DEBUG_PRINT(2, "DCF77: Local time valid.\n");
						time_valid = 1;
					}
				}
			}
		} else {
			DEBUG_PRINT(1, "DCF77: Datagram invalid.\n");
			time_valid = 0;
		}
	}

	if(minute_end || dgram_end) {
		/* Clear the second counter to enable receiving a new datagram: */
		dcf77_info.second = 0;
		dcf77_info.set_time = 0;
		DEBUG_PRINT(2, "DCF77: New datagram start.\n");

		/* If the datagram is too long or we received a partial datagram: */
		if(minute_end != dgram_end)
			time_valid = 0;

		if(!time_valid) {
			if(dcf77_info.invalid_frames < DCF77_MAX_INVALID_FRAMES)
				dcf77_info.invalid_frames++;
			DEBUG_PRINT(2, "DCF77: recv invalid frame (ctr %u).\n", dcf77_info.invalid_frames);
		} else
			dcf77_info.invalid_frames = 0;

		if(dcf77_info.invalid_frames >= DCF77_MAX_INVALID_FRAMES) {
			timer_info.valid = 0;
			DEBUG_PRINT(1, "DCF77: Max invalid frames.\n");
			dcf77_clear_sync();
		}
	}
}

uint8_t inline dcf77_gen_mask(uint8_t start, uint8_t len) {
	uint8_t mask = 0;
	uint8_t stop = start + len - 1;
	uint8_t i = 8, shiftmask = 128;
	if(len > 0) {
		do {
			i--;
			if(i <= stop)
				mask |= shiftmask;
			shiftmask >>= 1;
		} while(i > start);
	}
	return mask;
}

/*
 * dcf77_get_byte_bits:
 * Return a subset of the bits of one byte in a datagram.
 */
static struct dcf77_byte noinline dcf77_get_byte_bits(uint8_t byte,
	uint8_t mask) {
	struct dcf77_byte value;
	/* Determine if all selected bits are valid. If not, try the cumulative datagram. */
	uint8_t valid = (dcf77_datagram_valid[DCF77_STD_DGRAM][byte] & mask) == mask;
	if(valid) {
		value.data = dcf77_datagram[DCF77_STD_DGRAM][byte] & mask;
	} else {
		valid = (dcf77_datagram_valid[DCF77_CUMULATIVE_DGRAM][byte] & mask) == mask;
		valid *= 2;
		value.data = dcf77_datagram[DCF77_CUMULATIVE_DGRAM][byte] & mask;
	}
	value.valid = valid;
	return value;
}

/* Collect nr_bits from the datagram: */
static struct dcf77_byte dcf77_get_bits(uint8_t start, uint8_t nr_bits) {
	uint8_t byte = start / 8, len = start % 8, mask, inv_len = 8 - len;
	struct dcf77_byte value;

	mask = dcf77_gen_mask(len, nr_bits);

	value = dcf77_get_byte_bits(byte, mask);
	/* Lower the bits so that the first one sits at position 0: */
	value.data >>= len;

	/* Perhaps, wanted bits are in the second byte: */
	if(inv_len < nr_bits) {
		/* get the first nr_bits - len bytes from the next byte and stuff them before the old ones */
		len = nr_bits - inv_len;
		byte++;
		mask = pgm_read_byte(&dcf77_byte_mask[len]);
		struct dcf77_byte temp = dcf77_get_byte_bits(byte, mask);

		/* The valid state is 0 if one of both bytes was 0. Else, it is the maximum of both. */
		value.data |= temp.data << inv_len;
		value.valid *= temp.valid;
		if(value.valid > 2) value.valid = 2;
	}
	return value;
}

/*
 * dcf77_clear_datagrams:
 * Invalidate datagram to reflect a possible change in the transmitted data:
 * This occurs on DST, hour or day changes.
 */
void dcf77_clear_datagrams(void) {
	bzero(dcf77_datagram[DCF77_CUMULATIVE_DGRAM], 8);
	bzero(dcf77_datagram_valid[DCF77_CUMULATIVE_DGRAM], 8);
}

struct dcf77_entity {
	uint8_t second;
	uint8_t length:4;
	uint8_t check_parity:1;
};

union dcf77_entity_u {
	struct dcf77_entity e;
	uint16_t val;
};

struct dcf77_entity dcf77_entity_pos[] PROGMEM = {
	{0,	1, 0},
	{16, 1, 0},
	{17, 2, 0},
	{19, 1, 0},
	{20, 1, 0},
	{21, 7, 1},
	{29, 6, 1},
	{36, 6, 0},
	{42, 3, 0},
	{45, 5, 0},
	{50, 8, 1},
	{0,0,0} /* dummy entry to end the list */
	};

/*
 * dcf77_decode_datagram:
 * Put the received datagram bits together into a date_time structure.
 * Furthermore validate the received information.
 */
void dcf77_decode_datagram(void) {
		struct dcf77_byte bcd;
		uint8_t data, parity_valid;
		struct dcf77_entity *dentP = dcf77_entity_pos;
		union dcf77_entity_u dent;
		uint8_t entity_end;
		bcd.valid = 1;

		do {
			dent.val = (uint16_t) pgm_read_word(dentP);
			/*
			 * Match an entry if it is over.
			 */
			entity_end = dent.e.second + dent.e.length;
			uint8_t entity_total_len = entity_end + dent.e.check_parity;
			if(entity_total_len == dcf77_info.second) {
				bcd = dcf77_get_bits(dent.e.second, dent.e.length);
				data = dcf77_bcd2bin(bcd.data);
				if(dent.e.check_parity) {
					parity_valid = dcf77_check_parity(bcd.data,
						entity_end);
				}
				break;
			} else if(entity_total_len > dcf77_info.second) {
				entity_end = 0;
			}
			dentP++;
		} while(entity_end);

		/* Is there anything to do here? */
		if(!entity_end)
			return;

		switch(dcf77_info.second) {
		case 0+1:
			/* Datagram start. No valid information here. */
			if(data) bcd.valid = 0;
			dcf77_info.time_valid = bcd.valid > 0;
			break;

		case 16+1:
			if(bcd.valid)
				dcf77_info.expect_dst_change = data;
			DEBUG_PRINT(2, "DST change: %d, valid: %d\n", data, bcd.valid);
			break;

		case 17+2:
			if(data != 1 && data != 2)
				bcd.valid = 0;
			if(bcd.valid) timer_dcf77.dst = data == 1;
			DEBUG_PRINT(2, "MEZ(2)/MESZ(1): %d, valid: %d\n", data, bcd.valid);
			break;

		case 19+1:
			if(bcd.valid)
				dcf77_info.expect_leapsecond = data;
			DEBUG_PRINT(2, "Leap second: %d, valid: %d\n", data, bcd.valid);
			break;

		case 20+1:
			/* Begin of time information: */
			if(data != 1)
				bcd.valid = 0;
			break;

		case 21+7+1:
			/* Obviously, the minute changes every minute, so do not use cached information here. */
			if(parity_valid != 1) {
				DEBUG_PRINT(2, "-EPARITY %u\n", parity_valid);
				bcd.valid = 0;
			}
			timer_dcf77.minute = data;
			if(bcd.valid == 1 && data == 0) {
				/*
				 * Clear the cumulative datagram as the received datagram
				 * may change every hour (different hour, day, month, year etc.)
				 */
				dcf77_clear_datagrams();
				if(dcf77_info.expect_leapsecond) {
					timer_add_leapsecond();
					DEBUG_PRINT(1, "DCF77: Leap second this minute!\n");
				}
			}

			DEBUG_PRINT(2, "Minute: %d valid: %d\n", data, bcd.valid);
			break;

		case 29+6+1:
			if(!parity_valid) {
				bcd.valid = 0;
				DEBUG_PRINT(2, "-EPARITY %u\n", parity_valid);
			}
			timer_dcf77.hour = data;
			DEBUG_PRINT(2, "Hour: %d, valid: %d\n", data, bcd.valid);
			break;

		case 36+6:
			timer_dcf77.day = data;
			DEBUG_PRINT(2, "Day: %d, valid: %d\n", data, bcd.valid);
			break;

		case 42+3:
			timer_dcf77.dow = data % 7;
			DEBUG_PRINT(2, "DOW: %d, valid: %d\n", data, bcd.valid);
			break;

		case 45+5:
			timer_dcf77.month = data;
			DEBUG_PRINT(2, "Month: %d, valid: %d\n", data, bcd.valid);
			break;

		case 50+8+1:
			timer_dcf77.year = data + 2000;
			uint32_t8 date_bits;
			date_bits.b1 = dcf77_get_bits(36, 8).data;
			date_bits.b2 = dcf77_get_bits(44, 8).data;
			date_bits.b3 = dcf77_get_bits(52, 6).data;
			if(!dcf77_check_parity(date_bits.val, 58))
				bcd.valid = 0;
			if(!bcd.valid)
				dcf77_info.time_valid = 0;
			DEBUG_PRINT(2, "Year: %d, valid: %d\n", data, bcd.valid);
			break;
		}
		if(!bcd.valid)
			dcf77_info.time_valid = 0;
		DEBUG_PRINT(2, "DCF77: dgram valid: %d\n", dcf77_info.time_valid);
}


static uint8_t dcf77_check_parity(uint32_t value, uint8_t parbitno) {
	struct dcf77_byte bcd = dcf77_get_bits(parbitno, 1);
	uint8_t bitcounter = bcd.data, i = 32;
	uint32_t mask = 1;
	do {
		bitcounter += (value & mask) > 0;
		mask <<= 1;
	} while(--i);
	uint8_t valid = !(bitcounter % 2);
	return valid?bcd.valid:0;
}

/*
 * Set the dgram_bit'th bit in the dram_no'th element of dgram to
 * the supplied value:
 */
static void dcf77_set_dgram_bit(uint8_t dgram[][DGRAM_LEN], uint8_t dgram_no, uint8_t dgram_bit,
	uint8_t value) {

	uint8_t bitfield_offset = dgram_bit/8;
	/* Create an mask for the affected bit: */
	uint8_t bitfield = 1 << (dgram_bit % 8);
	uint8_t field = dgram[dgram_no][bitfield_offset];
	if(value) {
		/* Set the bit in the bitfield: */
		field |= bitfield;
	} else {
		/* Invert the bitfield */
		bitfield = ~bitfield;
		/* Clear the bit: */
		field &= bitfield;
	}
	dgram[dgram_no][bitfield_offset] = field;
}

/*
 * Tell the DCF77 module to resynchronize the clock:
 */
void dcf77_clear_sync(void) {
	dcf77_clear_edge_stats();
	dcf77_info.tick_stat_valid = 0;
	dcf77_info.timer_in_sync = 0;
	dcf77_info.set_time = 0;
	DEBUG_PRINT(1, "DCF77: Synchronization lost.\n");
}


/*
 * dcf77_process_stats:
 * Process statistical information of the DCF77 module.
 */
static void dcf77_process_stats(void) {
	if(mean_edge_counter >= DCF77_AVR_INTERVAL) {
		if(dcf77_info.time_valid && dcf77_info.timer_in_sync) {
			/* Simulate rounded integer division: */
			mean_edge_ticks += (mean_edge_counter + 1) / 2;
			mean_edge_ticks /= mean_edge_counter;
			int32_t diff;

			/*
			 * If there are less ticks than the delay, the clock is is too slow. -> diff < 0
			 * More ticks than the correction: The clock is too fast. -> diff > 0
			 */
			diff = mean_edge_ticks - dcf77_recv_delay;
			/* Reduce the difference or creating the correction to stabilize the loop: */
			diff *= 3;
			diff /= 4;
			if(diff < -65534)
				diff = -65534;
			if(diff > 65535)
				diff = 65535;

			DEBUG_PRINT(2, "DCF77: %lu Pulse delta: %u expected: %u diff: %i ticks\n",
				timer_unix_timestamp, timer_ctr_ticks2delay(mean_edge_ticks),
				timer_ctr_ticks2delay(dcf77_recv_delay), diff);
			if(diff) timer_skew_clock((int16_t) diff);
		}
		dcf77_clear_edge_stats();
	}
}

static noinline void dcf77_clear_edge_stats(void) {
	mean_edge_ticks = 0;
	mean_edge_counter = 0;
}

static inline uint8_t dcf77_pulse_ok(uint16_t level_length) {
	uint8_t ok;

	/* The end-of-minute pulse may be longer: */
	if(dcf77_info.second >= 58)
		ok = current_pulse_length < DCF77_MIN_LAST_PULSE_LENGTH;
	else
		ok = current_pulse_length < DCF77_MIN_PULSE_LENGTH;

	ok |= level_length < DCF77_MIN_DOWN_LENGTH;
	return ok;
}

uint8_t dcf77_is_minute_end(void) {
	timer_millistamp curr;
	timer_get_timestamp(&curr);
	uint16_t level_length = timer_diff_timestamp16(&dcf77_t_edge, &curr) + current_down_length;
	return level_length > DCF77_MAX_PULSE_LENGTH;
}

uint8_t noinline dcf77_seconds(void) {
	uint8_t second_top = 58;
	if(dcf77_info.expect_leapsecond && !timer_dcf77.minute) second_top = 59;
	return second_top;
}
