/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _DSEEPROM_LAYOUT_H
#define _DSEEPROM_LAYOUT_H

#include <stdlib.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include "config.h"
#include "dseeprom.h"

/*
 * EEPROM layout description:
 * The EEPROM has a dummy byte (it can be easily corrupted on old AVRs) at
 * the start, followed by a read-only default configuration.
 */


/* Includes for the stored element names: */
#include "uip/uip_arp.h"
#include "dcf77.h"
#include "timer.h"
#include "amish_uip.h"
#include "main.h"

#ifndef offsetof
	#define offsetof(t,m) (ptr_t)((&((t *)0L)->m))
#endif
#define sizeof_member(s, m) (sizeof(&((s *)0L)->m))

/*
 * These values are saved in the EEPROM:
 * Attention: the entries have to match their RAM counterpart's size.
 * See EEP_ELEM_ENTRY.size.
 */

struct eep_frame {
	uint8_t ethaddr[6];
	uint8_t ipv4addr[4];
	uint8_t ipv4mask[4];
	uint8_t ipv4gw[4];
	uint16_t dcf77_recv_delay;
	char amish_uip_pw[AMISH_PW_LEN+1];
	uint8_t display_brightness;

/* Internally used values: */
	uint8_t valid1; /* Valid values for valid1/2: 0x0, 0xFF */
	uint8_t valid2; /* A configuration is only valid if valid1 != valid2 */
	uint8_t crc8; /* CRC8 is calculated over eep_frame excluding this field. */
};

/* Mind a padding: */
#define EEP_FRAME_SIZE (sizeof(struct eep_frame) * E2PAGESIZE)

/*
 * Put every eep_frame value also into this list:
 */
#define EEP_ELEM_ENTRY(ram_name, eeprom_name, size_elem) \
	{(void*) ram_name, (uint16_t) offsetof(struct eep_frame, eeprom_name), sizeof(size_elem)}
struct {
	void *ram;
	uint8_t eeprom;
	uint8_t size;
} eep_elements[] PROGMEM = {
	EEP_ELEM_ENTRY(uip_ethaddr.addr, ethaddr, uip_ethaddr.addr),
	EEP_ELEM_ENTRY(uip_hostaddr, ipv4addr, uip_hostaddr),
	EEP_ELEM_ENTRY(uip_netmask, ipv4mask, uip_netmask),
	EEP_ELEM_ENTRY(uip_draddr, ipv4gw, uip_draddr),
	EEP_ELEM_ENTRY(&dcf77_recv_delay, dcf77_recv_delay, dcf77_recv_delay),
	EEP_ELEM_ENTRY(amish_uip_pw, amish_uip_pw, amish_uip_pw),
	EEP_ELEM_ENTRY(&display_brightness, display_brightness, display_brightness),
	{NULL, 0, 0}
};


/* Default values are loaded from this structure: */
struct {
	/*
	 * This value may be overwritten when the voltage drops during an EEPROM
	 * cell update:
	 */
	uint8_t unsafe_dummy[E2PAGESIZE];
	struct eep_frame defaults;
#define EEP_INIT_SIZE (E2PAGESIZE /* dummy */ + sizeof(struct eep_frame))
#define EEP_INIT_PAD_SIZE (EEP_INIT_SIZE % E2PAGESIZE)
	/* Warning: EEP_INIT_PAD_SIZE==0 only works with GCC! */
	uint8_t init_padding[EEP_INIT_PAD_SIZE];
	/* First address after the initialization block, on a new EEPROM block: */
#define EEP_CONFIG_SIZE (E2END+1 - (EEP_INIT_SIZE + EEP_INIT_PAD_SIZE + 4))
	uint8_t config_space[EEP_CONFIG_SIZE];
	uint8_t write_counter[4]; /* avrdude flash write counter */
} dseeprom_contents EEMEM =
{{0}, {
	NET_ETHER, /* ethaddr */
	NET_IP, /* ipv4addr */
	NET_MASK, /* ipv4mask */
	NET_GW, /*ipv4gw */
	(((uint32_t) DCF77_DELAY_COMPENSATION) * TIMER_SCALE_VALUE / 100000),
		/* dcf77recv_delay */
	AMISH_UIP_DEFAULT_PW, /* AMISH default password */
	1 /* Default display brightness: lowest */
},
};


#endif
