/*
 * Copyright (C) 2009-2010 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "avrutils.h"

#include "dseeprom.h"
#include "dseeprom-layout.h"

uint8_t eep_idx; /* A pointer to the current configuration index */

enum wstate {
	idle = 0, writing, update_valid, update_crc, invalidate_old
};

struct {
	enum wstate state:3;
	uint8_t repeat:1; /* Set to write a new config after a previous one */
} eep_winfo;

uint8_t eep_w_elem, eep_w_offset, eep_w_crc;

static void    eep_writer(void);
static uint8_t eep_load(void* start_address, uint8_t interleaved);
       void    eep_save(void);
       uint8_t eep_load_valid(void);

ISR(EE_READY_vect) {
	/*
	 * Disable the write completion interrupt as it fires continuously if
	 * EEWE is cleared. Meaning: Not checking if an write request was issued
	 * before. Stupid thingy.
	 */
	EECR &= ~(1 << EERIE);

	/* A write access finished, do the next one: */
	eep_writer();
}

void eep_init(void) {
	/* load first valid config. */
	if(!eep_load_valid()) {
		eep_force_init();
	}
}

uint8_t eep_load_valid(void) {
	uint8_t idx = 0;
	uint8_t *confp = (uint8_t*)&dseeprom_contents.config_space;
	do {
		if(eep_load(confp, 1)) {
			eep_idx = idx;
			DEBUG_PRINT(2, "EEP: Config %u loaded.\n", idx);
			return 1;
		}
		confp += EEP_FRAME_SIZE;
	} while (++idx < DSEEPROM_NUM_CONFS);
	INFO_PRINT("EEPROM: Invalid config, loading defaults");
	return 0;
}

void eep_force_init(void) {
	/* Load default values */
	eep_load(&dseeprom_contents.defaults, 0);

	/* Clear valid marks on all frames */
	eep_idx = DSEEPROM_NUM_CONFS - 1;
	uint8_t i = 0;
	uint8_t *eeaddr = (uint8_t*) &dseeprom_contents.config_space;
	eeaddr += offsetof(struct eep_frame, valid1) * E2PAGESIZE;
	/*
	 * The valid and CRC bytes occupy three consecutive EEPROM blocks, starting
	 * at the first valid1 byte.
	 */
	do {
		eeprom_write_byte(eeaddr, 0xFF);
		eeaddr += E2PAGESIZE;
		eeprom_write_byte(eeaddr, 0xFF);
		eeaddr += E2PAGESIZE;
		eeprom_write_byte(eeaddr, 0xFF);
		eeaddr -= 2*E2PAGESIZE;
		eeaddr += EEP_FRAME_SIZE;
	} while(++i < DSEEPROM_NUM_CONFS);


	/* Save values to new config */
	eep_save();
}

static inline void eep_writer(void) {
	/*
	 * eeprom_write_byte waits at the beginning for EEWE to be cleared.
	 * This should hold if we are called in an interrupt contect.
	 */
	// eeprom_write_byte()
	uint8_t *eep, *ramp, data, elem, offset, new_idx;
	enum wstate state = eep_winfo.state;
	new_idx = (eep_idx + 1) % DSEEPROM_NUM_CONFS;
	eep = (uint8_t*) &dseeprom_contents.config_space + new_idx*EEP_FRAME_SIZE;


	switch(state) {
	case idle:
	default:
		/* Nothing to do here */
		eep = NULL;
		break;

	case writing:
		elem = eep_w_elem; offset = eep_w_offset;
		ramp = (uint8_t*) pgm_read_word(&eep_elements[elem].ram);
		eep += pgm_read_byte(&eep_elements[elem].eeprom) * E2PAGESIZE;
		uint8_t len = (uint8_t)  pgm_read_word(&eep_elements[elem].size);
		eep += offset * E2PAGESIZE;
		data = *(ramp + offset);
		if(offset == 0 && elem == 0) eep_w_crc = 0;
		/* Update the CRC with the new value: */
		eep_w_crc = crc8(eep_w_crc, data);

		/* Check if this is the last byte of the current element: */
		if(++offset >= len) {
			elem++;
			if((void*)pgm_read_word(&eep_elements[elem].ram) == NULL) {
				/* No next element found. Next state, please. */
				state = update_valid;
			} else {
				/* Switch over to the next element. */
				offset = 0;
			}
		}
		/* Write back modified values: */
		eep_w_elem = elem; eep_w_offset = offset;
		break;

	case update_valid:
		/* Read the old valid2 value: */
		eep += offsetof(struct eep_frame, valid2) * E2PAGESIZE;
		data = eeprom_read_byte(eep);
		if(data > 0) /* valid1 = ~valid2: */
			data = 0;
		else
			data = 0xFF;
		eep -= E2PAGESIZE; /* Write to valid1, please */
		state = update_crc;
		break;

	case update_crc:
		eep += offsetof(struct eep_frame, crc8) * E2PAGESIZE;
		data = eep_w_crc;
		state = invalidate_old;
		break;

	case invalidate_old:
		/* Copy valid1 to valid2 of the current frame, marking it as invalid: */
		eep = (uint8_t*) &dseeprom_contents.config_space + eep_idx*EEP_FRAME_SIZE +
			offsetof(struct eep_frame, valid1) * E2PAGESIZE;
		data = eeprom_read_byte(eep);
		eep += E2PAGESIZE; /* valid 1 -> valid 2 */

		/* Switch to the new index: */
		eep_idx = new_idx;
		if(eep_winfo.repeat) {
			state = writing;
			eep_w_offset = 0;
			eep_w_elem = 0;
			eep_winfo.repeat = 0;
		} else
			state = idle;
		break;
	}

	/* Change the state and fire off the write command: */
	eep_winfo.state = state;
	if(eep) eeprom_write_byte(eep, data);
	/* Enable EEPROM write completion interrupt, if a write request was issued: */
	if(state) EECR |= (1 << EERIE);
}

/*
 * eep_load:
 * Load a configuration from the specified address
 * to the RAM.
 * If interleaved is set, the configuration is loaded with a padding
 * after each EEPROM byte.
 * Return value: 1, if CRC is valid, else: 0.
 */
static uint8_t eep_load(void* start_address, uint8_t interleaved) {
	uint8_t crc = 0, eep_crc = 0, valid1, valid2;
	uint8_t i = 0, j;

	/* First, check the frame validity bytes and load the stored CRC. */
	if(interleaved) {
		uint8_t *eeptr = start_address +
			offsetof(struct eep_frame, valid1) * E2PAGESIZE;
		valid1 = eeprom_read_byte(eeptr);
		eeptr += E2PAGESIZE;
		valid2 = eeprom_read_byte(eeptr);
		eeptr += E2PAGESIZE;
		eep_crc = eeprom_read_byte(eeptr);
		if(valid1 != 0 && valid1 != 0xFF)
			return 0;
		if(valid2 != 0 && valid2 != 0xFF)
			return 0;
		if(valid1 == valid2)
			return 0;
	}

	/* For each stored variable: */
	do {
		uint8_t *ramp = (uint8_t*) pgm_read_word(&eep_elements[i].ram);
		uint8_t *eepp = start_address;
		uint8_t offset = pgm_read_byte(&eep_elements[i].eeprom);
		if(interleaved)
			offset *= E2PAGESIZE;
		eepp +=	offset;
		uint8_t len   = (uint8_t)  pgm_read_word(&eep_elements[i].size);
		if(!ramp) break;

		/* Load the stored variable to their respective position in the RAM: */
		j = 0;
		do {
			uint8_t data = eeprom_read_byte(eepp);
			*ramp = data;
			crc = crc8(crc, data);
			ramp++;
			if(interleaved)
				eepp += E2PAGESIZE;
			else
				eepp++;
		} while(++j < len);
		i++;
	} while(1);
	return crc == eep_crc;
}


void eep_save(void) {
	ENTER_CRITICAL_SECTION
		if(eep_winfo.state == idle) {
			eep_winfo.repeat = 0;
			eep_w_elem = 0;
			eep_w_offset = 0;
			eep_winfo.state = writing;
			/* Call the writer by enabling the EEPROM ready interrupt: */
			EECR |= (1 << EERIE);
		} else {
			/* Repeat the writing process if it is already running: */
			eep_winfo.repeat = 1;
		}
	LEAVE_CRITICAL_SECTION
}
