/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _ENC0_H
#define _ENC0_H

#include <inttypes.h>
#include "config.h"
#include "uip/uip.h"
#include "uip/uip_arp.h"

/*
 * enc0_init:
 * Call this at the start to initialize the driver and hardware.
 */
extern void enc0_init(void);
/*
 * enc0_send_packet:
 * Send a packet via Ethernet, including ARP resolution beforehand.
 * Returns 0 if it succeeded, != 0 if an error occured.
 */
extern uint8_t enc0_send_packet(void);

/*
 * enc0_tx_frame:
 * Send a raw packet via Ethernet, without any change.
 * Returns 0 if it succeeded, != 0 if an error occured.
 */
uint8_t enc0_tx_frame(void);

/*
 * enc0_worker:
 * Worker function that processes RX + TX.
 */
void enc0_worker(void);

/* XXX: UIP IP/Ethernet address conversion, move to some other place */
extern char* enc0_sputip(char* dest, uip_ipaddr_t *ip);
extern char* enc0_sputmac(char* dest, struct uip_eth_addr *eth);
extern uint8_t enc0_sscanip(char* src, uip_ipaddr_t *ip);
extern uint8_t enc0_sscanmac(char* src, struct uip_eth_addr *eth);
/*
 * enc0_update_mac:
 * Update what the ENC28J60 hardware thinks is its MAC address.
 * Call this after updating uip_ethaddr.
 */
extern void enc0_update_mac(void);

/*
 * enc0_retransmit:
 * Retransmission helper.
 * XXX: Call this in your main loop, should be moved to dsev
 */
extern void enc0_retransmit(void);

#ifdef ENC0_INTERNAL

/* SPI opcodes: */
#define ENC0_OP_RCR	(0 << 5)
#define ENC0_OP_RBM ((1 << 5) | 26)
#define ENC0_OP_WCR	(2 << 5)
#define ENC0_OP_WBM ((3 << 5) | 26)
#define ENC0_OP_BFS	(4 << 5)
#define ENC0_OP_BFC	(5 << 5)
#define ENC0_OP_SRC ((7 << 5) | 31)

/*
 * Register list for ENC28J60:
 * The MAC and MII registers return a dummy byte first. ETH registers do not.
 * To further complicate the issue, there are 4 register banks.
 * As 3 bits are left in an address byte, organize these as follows:
 * BBEAAAAA -- BB is the bank number, E indicates an ETH register (no dummy
 * byte returned), and AAAAA is the register address.
 */

#define ENC0_REG_BANK_SHIFT 6
#define ENC0_REG_TYPE_SHIFT 5
#define ENC0_REG_BANK(bankno) ((bankno) << ENC0_REG_BANK_SHIFT)
#define ENC0_REG_TYPE(type) ((type) << ENC0_REG_TYPE_SHIFT)

/* Masks for the register address parts: */
#define ENC0_REG_ADDR_MASK 31
#define ENC0_REG_BANK_MASK ENC0_REG_BANK(3)
#define ENC0_REG_TYPE_MASK ENC0_REG_TYPE(1)

#define ENC0_REG_ETH (1 << ENC0_REG_TYPE_SHIFT)
#define ENC0_REG_MAC (0 << ENC0_REG_TYPE_SHIFT)
#define ENC0_REG_MII (0 << ENC0_REG_TYPE_SHIFT)


/* registers beginning at 0x1A are the same in every bank: */
#define ENC0_REG_IN_ALL_BANKS 0x1A

/* Bank 0 */
#define ERDPT		ERDPTL
#define ERDPTL	 ((uint8_t) (0x00 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERDPTH	 ((uint8_t) (0x01 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EWRPT		EWRPTL
#define EWRPTL	 ((uint8_t) (0x02 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EWRPTH	 ((uint8_t) (0x03 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ETXST		ETXSTL
#define ETXSTL	 ((uint8_t) (0x04 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ETXSTH	 ((uint8_t) (0x05 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ETXND		ETXNDL
#define ETXNDL	 ((uint8_t) (0x06 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ETXNDH	 ((uint8_t) (0x07 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXST		ERXSTL
#define ERXSTL	 ((uint8_t) (0x08 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXSTH	 ((uint8_t) (0x09 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXND		ERXNDL
#define ERXNDL	 ((uint8_t) (0x0A | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXNDH	 ((uint8_t) (0x0B | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXRDPT	ERXRDPTL
#define ERXRDPTL ((uint8_t) (0x0C | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXRDPTH ((uint8_t) (0x0D | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXWRPT	ERXWRPTL
#define ERXWRPTL ((uint8_t) (0x0E | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ERXWRPTH ((uint8_t) (0x0F | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMAST	 EDMASTL
#define EDMASTL	((uint8_t) (0x10 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMASTH	((uint8_t) (0x11 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMAND	 EDMANDL
#define EDMANDL	((uint8_t) (0x12 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMANDH	((uint8_t) (0x13 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMADST	EDMADSTL
#define EDMADSTL ((uint8_t) (0x14 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMADSTH ((uint8_t) (0x15 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMACS	 EDMACSL
#define EDMACSL	((uint8_t) (0x16 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EDMACSH	((uint8_t) (0x17 | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EIE			((uint8_t) (0x1B | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define EIR			((uint8_t) (0x1C | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ESTAT		((uint8_t) (0x1D | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ECON2		((uint8_t) (0x1E | ENC0_REG_BANK(0) | ENC0_REG_ETH))
#define ECON1		((uint8_t) (0x1F | ENC0_REG_BANK(0) | ENC0_REG_ETH))

/* Bank 1 */
#define EHT0		 ((uint8_t) (0x00 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT1		 ((uint8_t) (0x01 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT2		 ((uint8_t) (0x02 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT3		 ((uint8_t) (0x03 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT4		 ((uint8_t) (0x04 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT5		 ((uint8_t) (0x05 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT6		 ((uint8_t) (0x06 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EHT7		 ((uint8_t) (0x07 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM0		((uint8_t) (0x08 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM1		((uint8_t) (0x09 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM2		((uint8_t) (0x0A | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM3		((uint8_t) (0x0B | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM4		((uint8_t) (0x0C | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM5		((uint8_t) (0x0D | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM6		((uint8_t) (0x0E | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMM7		((uint8_t) (0x0F | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMCS		EPMCSL
#define EPMCSL	 ((uint8_t) (0x10 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMCSH	 ((uint8_t) (0x11 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMO		 EPMOL
#define EPMOL		((uint8_t) (0x14 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPMOH		((uint8_t) (0x15 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define ERXFCON	((uint8_t) (0x18 | ENC0_REG_BANK(1) | ENC0_REG_ETH))
#define EPKTCNT	((uint8_t) (0x19 | ENC0_REG_BANK(1) | ENC0_REG_ETH))

/* Bank 2 */
#define MACON1	 ((uint8_t) (0x00 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MACON3	 ((uint8_t) (0x02 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MACON4	 ((uint8_t) (0x03 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MABBIPG	((uint8_t) (0x04 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MAIPG		MAIPGL
#define MAIPGL	 ((uint8_t) (0x06 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MAIPGH	 ((uint8_t) (0x07 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MACLCON1 ((uint8_t) (0x08 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MACLCON2 ((uint8_t) (0x09 | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MAMXFL	 MAMXFLL
#define MAMXFLL	((uint8_t) (0x0A | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MAMXFLH	((uint8_t) (0x0B | ENC0_REG_BANK(2) | ENC0_REG_MAC))
#define MICMD		((uint8_t) (0x12 | ENC0_REG_BANK(2) | ENC0_REG_MII))
#define MIREGADR ((uint8_t) (0x14 | ENC0_REG_BANK(2) | ENC0_REG_MII))
#define MIWR		 MIWRL
#define MIWRL		((uint8_t) (0x16 | ENC0_REG_BANK(2) | ENC0_REG_MII))
#define MIWRH		((uint8_t) (0x17 | ENC0_REG_BANK(2) | ENC0_REG_MII))
#define MIRD		 MIRDL
#define MIRDL		((uint8_t) (0x18 | ENC0_REG_BANK(2) | ENC0_REG_MII))
#define MIRDH		((uint8_t) (0x19 | ENC0_REG_BANK(2) | ENC0_REG_MII))

/* Bank 3 */
#define MAADR5	 ((uint8_t) (0x00 | ENC0_REG_BANK(3) | ENC0_REG_MAC))
#define MAADR6	 ((uint8_t) (0x01 | ENC0_REG_BANK(3) | ENC0_REG_MAC))
#define MAADR3	 ((uint8_t) (0x02 | ENC0_REG_BANK(3) | ENC0_REG_MAC))
#define MAADR4	 ((uint8_t) (0x03 | ENC0_REG_BANK(3) | ENC0_REG_MAC))
#define MAADR1	 ((uint8_t) (0x04 | ENC0_REG_BANK(3) | ENC0_REG_MAC))
#define MAADR2	 ((uint8_t) (0x05 | ENC0_REG_BANK(3) | ENC0_REG_MAC))
#define EBSTSD	 ((uint8_t) (0x06 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define EBSTCON	((uint8_t) (0x07 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define EBSTCS	 EBSTCSL
#define EBSTCSL	((uint8_t) (0x08 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define EBSTCSH	((uint8_t) (0x09 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define MISTAT	 ((uint8_t) (0x0A | ENC0_REG_BANK(3) | ENC0_REG_MII))
#define EREVID	 ((uint8_t) (0x12 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define ECOCON	 ((uint8_t) (0x15 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define EFLOCON	((uint8_t) (0x17 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define EPAUS		EPAUSL
#define EPAUSL	 ((uint8_t) (0x18 | ENC0_REG_BANK(3) | ENC0_REG_ETH))
#define EPAUSH	 ((uint8_t) (0x19 | ENC0_REG_BANK(3) | ENC0_REG_ETH))

/* PHY */
#define PHCON1	0x00
#define PHSTAT1 0x01
#define PHID1	 0x02
#define PHID2	 0x03
#define PHCON2	0x10
#define PHSTAT2 0x11
#define PHIE		0x12
#define PHIR		0x13
#define PHLCON	0x14


/* Values of the registers: */

/* Bank 0 */
#define EIE_INTIE			 (1 << 7)
#define EIE_PKTIE			 (1 << 6)
#define EIE_DMAIE			 (1 << 5)
#define EIE_LINKIE			(1 << 4)
#define EIE_TXIE				(1 << 3)
#define EIE_TXERIE			(1 << 1)
#define EIE_RXERIE			(1 << 0)

#define EIR_PKTIF			 (1 << 6)
#define EIR_DMAIF			 (1 << 5)
#define EIR_LINKIF			(1 << 4)
#define EIR_TXIF				(1 << 3)
#define EIR_TXERIF			(1 << 1)
#define EIR_RXERIF			(1 << 0)

#define ESTAT_INT			 (1 << 7)
#define ESTAT_BUFER		 (1 << 6)
#define ESTAT_LATECO		ESTAT_LATECOL
#define ESTAT_LATECOL	 (1 << 4)
#define ESTAT_RXBUSY		(1 << 2)
#define ESTAT_TXABRT		(1 << 1)
#define ESTAT_CLKRDY		(1 << 0)

#define ECON2_AUTOINC	 (1 << 7)
#define ECON2_PKTDEC		(1 << 6)
#define ECON2_PWRSV		 (1 << 5)
#define ECON2_VRPS			(1 << 3)

#define ECON1_TXRST		 (1 << 7)
#define ECON1_RXRST		 (1 << 6)
#define ECON1_DMAST		 (1 << 5)
#define ECON1_CSUMEN		(1 << 4)
#define ECON1_TXRTS		 (1 << 3)
#define ECON1_RXEN			(1 << 2)
#define ECON1_BSEL1		 (1 << 1)
#define ECON1_BSEL0		 (1 << 0)

/* Bank 1 */
#define ERXFCON_UCEN		(1 << 7)
#define ERXFCON_ANDOR	 (1 << 6)
#define ERXFCON_CRCEN	 (1 << 5)
#define ERXFCON_PMEN		(1 << 4)
#define ERXFCON_MPEN		(1 << 3)
#define ERXFCON_HTEN		(1 << 2)
#define ERXFCON_MCEN		(1 << 1)
#define ERXFCON_BCEN		(1 << 0)

/* Bank 2 */
#define MACON1_TXPAUS	 (1 << 3)
#define MACON1_RXPAUS	 (1 << 2)
#define MACON1_PASSAL	 MACON1_PASSALL
#define MACON1_PASSALL	(1 << 1)
#define MACON1_MARXEN	 (1 << 0)

#define MACON3_PADCFG2	(1 << 7)
#define MACON3_PADCFG1	(1 << 6)
#define MACON3_PADCFG0	(1 << 5)
#define MACON3_TXCRCEN	(1 << 4)
#define MACON3_PHDREN	 (1 << 3)
#define MACON3_HFRMEN	 (1 << 2)
#define MACON3_FRMLNEN	(1 << 1)
#define MACON3_FULDPX	 (1 << 0)

#define MACON4_DEFER		(1 << 6)
#define MACON4_BPEN		 (1 << 5)
#define MACON4_NOBKOFF	(1 << 4)

#define MICMD_MIISCAN	 (1 << 1)
#define MICMD_MIIRD		 (1 << 0)

/* Bank 3 */
#define EBSTCON_PSV2		(1 << 7)
#define EBSTCON_PSV1		(1 << 6)
#define EBSTCON_PSV0		(1 << 5)
#define EBSTCON_PSE		 EBSTCON_PSEL
#define EBSTCON_PSEL		(1 << 4)
#define EBSTCON_TMSEL1	(1 << 3)
#define EBSTCON_TMSEL0	(1 << 2)
#define EBSTCON_TME		 (1 << 1)
#define EBSTCON_BISTST	(1 << 0)

#define MISTAT_NVALID	 (1 << 2)
#define MISTAT_SCAN		 (1 << 1)
#define MISTAT_BUSY		 (1 << 0)

#define ECOCON_COCON2	 (1 << 2)
#define ECOCON_COCON1	 (1 << 1)
#define ECOCON_COCON0	 (1 << 0)

#define EFLOCON_FULDPXS (1 << 2)
#define EFLOCON_FCEN1	 (1 << 1)
#define EFLOCON_FCEN0	 (1 << 0)

/* PHY registers */
#define PHCON1_PRST		 (1 << 15)
#define PHCON1_PLOOPBK	(1 << 14)
#define PHCON1_PPWRSV	 (1 << 11)
#define PHCON1_PDPXMD	 (1 << 8)

#define PHSTAT1_PFDPX	 (1 << 12)
#define PHSTAT1_PHDPX	 (1 << 11)
#define PHSTAT1_LLSTAT	(1 << 2)
#define PHSTAT1_JBSTAT	(1 << 1)

#define PHID2_PID24		 (1 << 15)
#define PHID2_PID23		 (1 << 14)
#define PHID2_PID22		 (1 << 13)
#define PHID2_PID21		 (1 << 12)
#define PHID2_PID20		 (1 << 11)
#define PHID2_PID19		 (1 << 10)
#define PHID2_PPN5			(1 << 9)
#define PHID2_PPN4			(1 << 8)
#define PHID2_PPN3			(1 << 7)
#define PHID2_PPN2			(1 << 6)
#define PHID2_PPN1			(1 << 5)
#define PHID2_PPN0			(1 << 4)
#define PHID2_PREV3		 (1 << 3)
#define PHID2_PREV2		 (1 << 2)
#define PHID2_PREV1		 (1 << 1)
#define PHID2_PREV0		 (1 << 0)

#define PHCON2_FRCLNK	 (1 << 14)
#define PHCON2_TXDIS		(1 << 13)
#define PHCON2_JABBER	 (1 << 10)
#define PHCON2_HDLDIS	 (1 << 8)

#define PHSTAT2_TXSTAT	(1 << 13)
#define PHSTAT2_RXSTAT	(1 << 12)
#define PHSTAT2_COLSTAT (1 << 11)
#define PHSTAT2_LSTAT	 (1 << 10)
#define PHSTAT2_DPXSTAT (1 << 9)
#define PHSTAT2_PLRITY	(1 << 5)

#define PHIE_PLNKIE		 (1 << 4)
#define PHIE_PGEIE			(1 << 1)

#define PHIR_PLNKIF		 (1 << 4)
#define PHIR_PGIF			 (1 << 2)

#define PHLCON_LACFG3	 (1 << 11)
#define PHLCON_LACFG2	 (1 << 10)
#define PHLCON_LACFG1	 (1 << 9)
#define PHLCON_LACFG0	 (1 << 8)
#define PHLCON_LBCFG3	 (1 << 7)
#define PHLCON_LBCFG2	 (1 << 6)
#define PHLCON_LBCFG1	 (1 << 5)
#define PHLCON_LBCFG0	 (1 << 4)
#define PHLCON_LFRQ1		(1 << 3)
#define PHLCON_LFRQ0		(1 << 2)
#define PHLCON_STRCH		(1 << 1)


/* Buffer memory management
 *
 * Split the memory into 2 equally sized chunks for RX/TX.
 * Follow the Errata sheet and use the lower half for RX.
 * Ensure below to use only odd addresses for	ERXRDPT.
 */
#define ENC0_RXBUF_START 0x0000 /* Receive buffer starts at the bottom of the memory. */
#define ENC0_RXBUF_END	 0x0FFF /* Receive buffer ends on an odd address */
#define ENC0_TXBUF_START 0x1000 /* Send buffer starts at half of the memory. */
#define ENC0_TXBUF_END	 0x1FFF /* The controller is initialized with this value. */


typedef union {
	uint8_t val[4];
	struct {
		uint16_t			packet_length;
		uint8_t prev_ignored:1;
		uint8_t rxdc_prev_seen:1;
		uint8_t carrier_prev_seen:1;
		uint8_t code_violation:1;
		uint8_t crc_error:1;
		uint8_t length_check_error:1;
		uint8_t length_out_of_range:1;
		uint8_t receive_ok:1;
		uint8_t mcast:1;
		uint8_t bcast:1;
		uint8_t dribble_nibble:1;
		uint8_t control_frame:1;
		uint8_t pause_control_frame:1;
		uint8_t unknown_opcode:1;
		uint8_t vlan_frame:1;
		uint8_t unused:1;
	};
} enc0_rx_header;

typedef union {
	uint8_t v[7];
	struct {
		uint16_t byte_count;
		uint8_t collision_count:4;
		uint8_t crc_error:1;
		uint8_t length_check_error:1;
		uint8_t length_too_large:1;
		uint8_t done:1;
		uint8_t multicast:1;
		uint8_t broadcast:1;
		uint8_t deferred:1;
		uint8_t excess_defer:1;
		uint8_t excess_coll:1;
		uint8_t late_coll:1;
		uint8_t giant:1;
		uint8_t underrun:1;
		uint16_t wire_transmitted;
		uint8_t control_frame:1;
		uint8_t pause_control_frame:1;
		uint8_t backpressure_applied:1;
		uint8_t VLANtagged:1;
		uint8_t unused:4;
	};
} enc0_tx_header;

#endif

#endif /* ENC0_INTERNAL */
