/*
 * Copyright (C) 2009-2011 Dennis Schneider
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *		notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *		notice, this list of conditions and the following disclaimer in the
 *		documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <avr/wdt.h>
#include <avr/io.h>
#include <stdio.h>
#include <util/delay.h>
#include <util/atomic.h>
#include "config.h"
#include "timer.h"
#include "dcf77.h"
#include "enc0.h"
#include "uip/uip.h"
#include "uip/uip_arp.h"
#include "uip-appcall.h"
#include "amish.h"
#include "dseeprom.h"
#include "twi.h"
#include "saa1064.h"
#include "main.h"
#include "avrutils.h"

#ifndef F_CPU
#error Please define the CPU frequency F_CPU
#endif

#define BUTTON_DEBOUNCE_DELEAY 10000 /* microseconds */

#define DISPLAY_CONF_LONGPRESS 10 /* deciseconds */

#define S_METER_LEVEL_DISCONNECTED 144 /* -dBm */

enum display_modes {
	DISPLAY_LOCALTIME = 0, DISPLAY_UTC, DISPLAY_DATE, DISPLAY_LEVEL
};

enum button_modes {
	BUTTON_RELEASED = 0, BUTTON_CHECK, BUTTON_PRESSED, BUTTON_LONG_PRESSED
};

struct {
	enum display_modes display_mode:2;
	uint8_t update_dp:1;
	uint8_t update_display:1;
	uint8_t was_locked:1;
	uint8_t button_pressed:1;
	enum button_modes button_mode:2;
	uint8_t old_brightness:4;
	uint8_t display_status_on:1;
} main_info;


uint8_t s_meter_level;
uint16_t pll_phase;
uint8_t supply_voltage;
uint16_t8 display_shadow[3];
uint16_t8 display_buffer[3];
uint8_t display_dps;
uint16_t button_pressed_counter;
stamp_t button_pressed_tick;
uint8_t display_brightness;

void dac_init(void);
void dac_process(void);
void lock_process(void);
void lock_clear_dcf77_sync(void);
uint16_t calc_s_meter_level(uint16_t u_adc);
void display_init(void);
void display_process(void);
void button_process(void);
void display_fill_buffer(uint8_t buffer_idx, uint8_t val);
void display_change_mode(void);
void display_update(void);
void display_level_update(void);
void display_blinker(void);
void display_config(void);
void display_config_ip(uint8_t intro, uip_ipaddr_t *addr);
uint8_t display_config_read8(uint8_t start);
uint8_t display_config_query(void);
void display_byte(uint8_t);

int main(void) {
	/*	Forcefully disable the watchdog. Else we end in an endless reset loop. */
#ifdef MCUCSR
	MCUCSR = 0;
#elif defined(MCUSR)
	MCUSR = 0;
#endif
	wdt_disable();

	/* Enable interrupts as early as possible, amish_uart needs it: */
	sei();

	amish_init();

	/* Port A is only used as input */
	DDRA = 0;
	/* Disable the analog comparator to save power: */
	ACSR = _BV(ACD);

	INFO_PRINT("\n\n"
		"DCF77 precision clock firmware\nVersion " FQ_VERSION
		"\n(C) 2009-2010 Dennis Schneider\n"
		"Contains code written by:\n"
		"\tAdam Dunkels, Fabian Greif, Florian Kristen, Joerg Wunsch,\n"
		"\tRoboterclub Aachen e.V., Swedish Institute of Computer Science\n");

	/* Initialize the submodules in a senseful order: */
	dac_init();
	eep_init();
	enc0_init();
	uip_arp_init();
	uip_init();
	timer_init();
	dcf77_init();
	uip_app_init();
	twi_init();
	display_init();


	/* Interactively configure the addresses: */
	display_config();

	/* Main program */
	while(1) {
		/*
		 * DCF77 edge detection. May be run more often, but it doesn't matter
		 * that much. The signal jitter is much larger than the time to run
		 * through this loop.
		 */
		dcf77_worker();

		/* These tick with 10 Hz: */
		if(timer_info.dtick) {
			timer_info.dtick = 0;
			enc0_retransmit();
			/* UIP periodic tasks are polled at 10 Hz. This should suffice. */
			uip_app_periodic();

			display_level_update();
			display_blinker();
		}

		/* Tick with 1 Hz: */
		if(timer_info.tick == 1) {
			timer_info.tick = 0;

			display_update(); /* Load the new time into the buffer */

			/* ARP cache maintenance every 60 seconds. */
			if(timer_localtime.second) uip_arp_timer();
			lock_process(); /* Has the PLL lock state changed? */

			/* If no antenna is connected, mark us as async. */
			if(s_meter_level == S_METER_LEVEL_DISCONNECTED)
				lock_clear_dcf77_sync();
		}


		amish_worker();

		enc0_worker();

		button_process(); /* Handle button events */
		display_process(); /* Write out the display's buffer */

		/* Permanently poll the ADC: */
		if(!(ADCSRA & _BV(ADSC))) dac_process();
	}
}


/*
 * Check if the PLL lock state has changed.
 * The clock only runs precise if the PLL is locked.
 */

void lock_process(void) {
	#define LOCKED	 0
	#define UNLOCKED 1
	/* Has the lock state changed? */
	uint8_t lock_state = IS_SET(PIN_LOCKED);

#if DEBUG_LEVEL
	uint8_t was_locked = main_info.was_locked;
	if(lock_state == was_locked)
		return;

	main_info.was_locked = lock_state;

	time_t t;
	ENTER_CRITICAL_SECTION
		t = timer_unix_timestamp;
	LEAVE_CRITICAL_SECTION
	DEBUG_PRINT(1, "%lu PLL state: %Slocked\n", t,
		lock_state == LOCKED ? PSTR("") : PSTR("un"));
#endif

	/*
	 * If we lost the lock, mark the clock as invalid to prevent
	 * NTP clients from synchronizing to it. Further, mark the current
	 * datagram as invalid to prevent the clock from synchronizing itself.
	 * And if we regain the lock, the clock synchronizes itself again.
	 */
	if(lock_state == UNLOCKED) {
		lock_clear_dcf77_sync();
	}
}

void lock_clear_dcf77_sync(void) {
		timer_info.valid = 0;
		dcf77_info.time_valid = 0;
		dcf77_info.timer_in_sync = 0;
}

#define DAC_U_REF 5
#define DAC_STEPS 1024
void dac_init(void) {
	/* Enable the ADC and set the prescaler to 64 or 128. */
	ADCSRA = (1<<ADEN) |
#if F_CPU == 10000000
		(1<<ADPS2) | (1<<ADPS1) | (0<<ADPS0); /* Prescaler 64 */
#elif F_CPU == 16000000
		(1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0); /* Prescaler 128 */
#else
	#error "dac_init: Unknown uC clock."
#endif
}

void dac_process(void) {
	enum dacs {
		S_METER = 0,
		PHASE = 1,
		SUPPLY_VOLTAGE = 2
	};
	static enum dacs dacsS;
	uint8_t dacno;
	enum dacs dacs = dacsS;

	/*
	 * The first result will be garbage, but we don't care.
	 * Subsequent calls produce the correct result because the
	 * ADC finishes within a millisecond.
	 */
	uint16_t adc_val = ADCW;

	static uint8_t adc_conv_count = 0;
	static uint32_t accu = 0;
	if(main_info.display_mode != DISPLAY_LEVEL) {
		accu += adc_val;
		if(!(++adc_conv_count)) {
			adc_conv_count = 0;
			adc_val = accu / 256;
			accu = 0;
		} else {
			ADCSRA |= (1<<ADSC);
			return;
		}
	} else {
		accu = 0;
		adc_conv_count = 0;
	}

	/* Save the digitized value and start the next conversion. */
	switch(dacs) {
	case S_METER:
		s_meter_level = calc_s_meter_level(adc_val);
		dacno = ADC_PHASE;
		break;

	case PHASE:
		pll_phase = adc_val;
		dacno = ADC_SUPPLY_VOLTAGE;
		break;

	case SUPPLY_VOLTAGE:
	default:
		/* The supply voltage is saved in decivolts. */
		#define SUPPLY_PRESCALER 4
		#define SUPPLY_MULTIPLIER 10
		supply_voltage = adc_val * 64 /
			(uint16_t)(DAC_STEPS * 64.0 / (SUPPLY_PRESCALER * DAC_U_REF * SUPPLY_MULTIPLIER));
		dacno = ADC_S_METER;
		break;
	}
	dacs++;
	if(dacs == 3) dacs = 0;

	/* If the signal level is displayed, update only this variable. */
	if(main_info.display_mode == DISPLAY_LEVEL) {
		dacno = ADC_S_METER;
		dacs = S_METER;
	}

	dacsS = dacs;

	ADMUX = dacno;
	/*
	 * ADMUX: REFSx both cleared so that AREF is used.
	 *				ADLAR cleared, result is right-justified.
	 */

	/* Start the conversion: */
	ADCSRA |= (1<<ADSC);

}

/*
 * Convert the signal strength voltage to a normalized
 * signal strength in -dBm.
 * Linear approximation calc_s_meter_level(u_adc) = a * u_adc + b
 * with a = -395/23, b = 6677/46
 * Reference voltage: 5V, so u_adc in volts is u_adc * 5 / 1024.
 * All operations are done with a multiplier of 1024 for more precision.
 */
uint16_t calc_s_meter_level(uint16_t u_adc) {
	uint32_t u_e = u_adc;
	u_e *= 1975 * 1024L;
	u_e /= 23552;
	return (uint16_t)((148636L - u_e)/1024);
}

void display_set_brightness(void) {
	uint8_t i = 0, bright = display_brightness;
	do {
		saa1064_set_brightness(i, bright);
	} while(++i < 3);
}

void display_change_brightness(void) {
	uint8_t bright = display_brightness;
	bright++;
	if(bright > DISPLAY_MAX_BRIGHTNESS) bright = 1;
	display_brightness = bright;
	/* Set the new brightness and save the result: */
	display_set_brightness();
}

void display_init(void) {
		saa1064_init(0);
		saa1064_init(1);
		saa1064_init(2);
		display_set_brightness();

		/* Select on the local time mode: */
		main_info.display_mode = DISPLAY_LEVEL;
		display_change_mode();
}

void display_process(void) {
	/* Only update the display if something has changed: */
	if(!main_info.update_display) return;
	main_info.update_display = 0;
	uint8_t force_change = main_info.update_dp;
	main_info.update_dp = 0;

	for(uint8_t i = 0; i < 3; i++) {
		uint16_t8 buffer;
		buffer.val = display_buffer[i].val;
		if(buffer.val == display_shadow[i].val && !force_change) continue;
		display_shadow[i] = buffer;
		/* This segment pair has changed. */

		if(main_info.display_status_on) {
			uint8_t mask = _BV(2*i);
			if(display_dps & mask) buffer.lo |= _BV(7);
			mask <<= 1;
			if(display_dps & mask) buffer.hi |= _BV(7);
		}

		saa1064_display(i, buffer);
	}
}

void button_process(void) {
	enum button_modes mode = main_info.button_mode;
	uint16_t counter;
	stamp_t stamp;

	switch(mode) {
	case BUTTON_RELEASED:
		if(!IS_SET(FN_BUTTON)) break;
		mode = BUTTON_CHECK;
		uint16_t next_tick;
		timer_get_ctr(next_tick);
		timer_ctr_add_ticks(next_tick, timer_ctr_us2ticks(BUTTON_DEBOUNCE_DELEAY));
		button_pressed_counter = next_tick;
		ENTER_CRITICAL_SECTION
			button_pressed_tick = timer_timestamp + 5;
		LEAVE_CRITICAL_SECTION
		break;

	case BUTTON_CHECK:
		counter = TCNT1;
		if(counter < button_pressed_counter
			|| counter - button_pressed_counter >
				timer_ctr_us2ticks(BUTTON_DEBOUNCE_DELEAY))
					break;
		if(!IS_SET(FN_BUTTON)) { mode = BUTTON_RELEASED; break;}
		main_info.old_brightness = display_brightness;
		mode = BUTTON_PRESSED;
		break;

	case BUTTON_PRESSED:
	case BUTTON_LONG_PRESSED:
		ENTER_CRITICAL_SECTION
			stamp = timer_timestamp;
		LEAVE_CRITICAL_SECTION
		if(!IS_SET(FN_BUTTON)) {
			if(mode == BUTTON_PRESSED)
				display_change_mode();
			else if(display_brightness != main_info.old_brightness)
				eep_save();
			mode = BUTTON_RELEASED;
		} else if (button_pressed_tick > stamp) {
			break;
		} else {
			display_change_brightness();
			mode = BUTTON_LONG_PRESSED;
		}
		button_pressed_tick = stamp + 5;
	}
	main_info.button_mode = mode;
}

noinline void display_fill_buffer(uint8_t buffer_idx, uint8_t val) {
	display_buffer[buffer_idx].lo = '0' + val % 10;
	display_buffer[buffer_idx].hi = '0' + val / 10;
}

void display_change_mode(void) {
	enum display_modes mode = main_info.display_mode;
	mode++;
	if(mode > DISPLAY_LEVEL) mode = DISPLAY_LOCALTIME;
	main_info.display_mode = mode;
	main_info.update_dp = 1;
	main_info.update_display = 1;

	/* Calculate the decimal points: */
	uint8_t dp;
	switch(mode) {
	case DISPLAY_LOCALTIME:
		dp = _BV(2 * LOCALTIME_ADDR + LOCALTIME_PORT);
		break;
	case DISPLAY_UTC:
		dp = _BV(2 * UTC_ADDR + UTC_PORT);
		break;
	case DISPLAY_DATE:
		dp = _BV(2 * DATE_ADDR + DATE_PORT);
		break;
	case DISPLAY_LEVEL:
	default:
		dp = 0;
		display_buffer[2].hi = 'P';
		display_buffer[2].lo = ' ';
	}
	display_dps = dp;
	display_update();
	display_level_update();
}

void display_update(void) {
	date_time d;
	timer_get_time(&d);

	uint8_t mode = main_info.display_mode;
	if(dcf77_info.first_set) {
		switch(mode) {
		case DISPLAY_LOCALTIME:
		case DISPLAY_UTC:
			display_fill_buffer(0, d.second);
			display_fill_buffer(1, d.minute);
			uint8_t hour = d.hour;
			if(main_info.display_mode == DISPLAY_UTC) {
				hour -= TIMER_TZ/60;
				if(timer_localtime.dst) hour--;
				if(hour >= 24)
#if					TIMER_TZ < 0
						hour -= 24;
#else
						hour += 24;
#endif
			}
			display_fill_buffer(2, hour);
			break;
		case DISPLAY_DATE:
			display_fill_buffer(0, d.year%100);
			display_fill_buffer(1, d.month);
			display_fill_buffer(2, d.day);
		}
	} else {
		uint8_t second;
		switch(mode) {
		case DISPLAY_DATE:
			display_change_mode();
			break;
		case DISPLAY_LOCALTIME:
		case DISPLAY_UTC:
			second = dcf77_info.second;
			if(dcf77_is_minute_end()) second = 59;
			if(mode != DISPLAY_DATE) display_fill_buffer(0, second);
			display_buffer[1].lo = 0;
			display_buffer[1].hi = 0;
			display_buffer[2].lo = 0;
			display_buffer[2].hi = 0;

			/* hack: Fake a second change to update the display faster */
			timer_info.tick = 1;
		}
	}
	main_info.update_display = 1;
}

void display_level_update(void) {
	if(main_info.display_mode != DISPLAY_LEVEL) return;
	main_info.update_display = 1;

	uint16_t level = s_meter_level;
	display_byte(level);
	if(level >= 100)
		display_buffer[1].hi = '-';
	else {
		display_buffer[1].hi = ' ';
		display_buffer[1].lo = '-';
	}

}

void display_blinker(void) {
	uint8_t ds = timer_info.decisecond_counter;
	if(ds != 5 && ds != 0) return;
	uint8_t state = main_info.display_status_on;
	if(dcf77_info.timer_in_sync) {
		if(state) return;
		state = 1;
		main_info.update_dp = 1;
		main_info.update_display = 1;
	} else {
		state = ds == 0;
		main_info.update_dp = 1;
		main_info.update_display = 1;
	}
	main_info.display_status_on = state;
}

noinline void display_forced_process(void) {
	main_info.update_display = 1;
	display_process();
}

void display_config(void) {
	uip_ipaddr_t addr;
	uint8_t changed = 0;

	display_buffer[2].hi = 'C';
	display_buffer[2].lo = 'F';
	display_buffer[1].hi = 'G';
	display_buffer[1].lo = '?';
	display_buffer[0].hi = ' ';
	display_buffer[0].lo = ' ';
	display_forced_process();
	uint8_t i = 30;
	do {
		_delay_ms(100);
		/* Only configure if the button is pressed. */
		if(IS_SET(FN_BUTTON)) goto do_config;
	} while(--i);
	return;
	do_config:

	display_buffer[1].lo = ' ';
	display_forced_process();

	while(IS_SET(FN_BUTTON));

	/* Change IP: */
	uip_ipaddr_copy(addr, uip_hostaddr);
	display_config_ip('I', &uip_hostaddr);
	changed = !uip_ipaddr_cmp(&addr, &uip_hostaddr);

	/* Change netmask: */
	uip_ipaddr_copy(addr, uip_netmask);
	display_config_ip('/', &uip_netmask);
	if(!uip_ipaddr_cmp(&addr, &uip_netmask)) changed = 1;

	/* Change gateway: */
	uip_ipaddr_copy(addr, uip_draddr);
	display_config_ip('G', &uip_draddr);
	if(!uip_ipaddr_cmp(&addr, &uip_draddr)) changed = 1;

	/* Ask if the password should be cleared: */
	display_buffer[2].hi = 'C';
	display_buffer[2].lo = 'L';
	display_buffer[1].hi = 'R';
	display_buffer[1].lo = 'P';
	display_buffer[0].hi = '?';
	display_buffer[0].lo = ' ';
	display_forced_process();
	if(display_config_query()) {
		amish_uip_pw[0] = '-';
		amish_uip_pw[1] = '\0';
		changed = 1;
	}

	/* Ask the user if the new settings should be stored: */
	if(changed) {
		display_buffer[2].hi = 'S';
		display_buffer[2].lo = 'T';
		display_buffer[1].hi = 'O';
		display_buffer[1].lo = 'R';
		display_buffer[0].hi = 'E';
		display_buffer[0].lo = '?';
	} else {
		display_buffer[2].hi = 'B';
		display_buffer[2].lo = 'E';
		display_buffer[1].hi = 'F';
		display_buffer[1].lo = 'O';
		display_buffer[0].hi = 'R';
		display_buffer[0].lo = 'E';
	}
	display_forced_process();
	while(IS_SET(FN_BUTTON));
	if(changed) eep_save();
}

void display_config_ip(uint8_t intro, uip_ipaddr_t *addr) {
	display_buffer[2].hi = intro;

	uint32_t8 val; val.val = *((uint32_t*) addr);

	uint8_t i = 0;
	do {
		display_buffer[2].lo = '1' + i;
		val.b[i] = display_config_read8(val.b[i]);
	} while(++i < 4);
	*((uint32_t*) addr) = val.val;
}

uint8_t display_config_read8(uint8_t value) {
	enum {
		RELEASED = 0, PRESSED, LONG_PRESSED
	} button_state = RELEASED;
	stamp_t s;
	uint8_t digit = 2, buffer[3];

	display_byte(value);

	buffer[2] = display_buffer[1].lo;
	buffer[1] = display_buffer[0].hi;
	buffer[0] = display_buffer[0].lo;

	/*
	 * If the button is pressed, move to the correct state:
	 * Releasing the button soon won't do anything, pressing longer
	 * changes the digit. As expected.
	 */
	if(IS_SET(FN_BUTTON)) {
		ENTER_CRITICAL_SECTION
			s = timer_timestamp + DISPLAY_CONF_LONGPRESS;
		LEAVE_CRITICAL_SECTION
		button_state = LONG_PRESSED;
	}

	while(1) {
		stamp_t ts;
		ENTER_CRITICAL_SECTION
			ts = timer_timestamp;
		LEAVE_CRITICAL_SECTION
		if(ts % 4 == 0) {
			((uint8_t*)display_buffer)[digit] = 0;
		}

		/* Force display update: */
		display_forced_process();

		switch(button_state) {
		case RELEASED:
			if(!IS_SET(FN_BUTTON)) break;
			s = ts + DISPLAY_CONF_LONGPRESS;

			_delay_ms(10);
			if(!IS_SET(FN_BUTTON)) break;

			button_state = PRESSED;
			break;

		case PRESSED:
		case LONG_PRESSED:
			if(!IS_SET(FN_BUTTON)) {
				if(button_state == PRESSED) { /* Short press: Increase digit */
					uint8_t number = buffer[digit];
					number++;
					if(number > '9') number = '0';
					if(digit == 2 && number > '2') number = '0';
					buffer[digit] = number;
				} /* else:	Long press quit, nothing to do */
				button_state = BUTTON_RELEASED;
			} else if (s > ts) {	/* Press still too short for a long press */
				break;
			} else { /* Long press, change digit. */
				if(digit == 0) goto read8_end;
				digit--;
				button_state = LONG_PRESSED;
			}
			s = ts + DISPLAY_CONF_LONGPRESS;
		}

		display_buffer[1].hi = ' ';
		display_buffer[1].lo = buffer[2];
		display_buffer[0].hi = buffer[1];
		display_buffer[0].lo = buffer[0];
	}
read8_end:
	value = buffer[0] - '0';
	value += 10 * (buffer[1] - '0');
	value += 100 * (buffer[2] - '0');
	return value;
}

void display_byte(uint8_t val) {
	display_fill_buffer(1, val / 100);
	display_buffer[1].hi = ' ';
	display_fill_buffer(0, val % 100);
}

uint8_t display_config_query(void) {
	enum {
		RELEASED = 0, PRESSED, LONG_PRESSED
	} button_state = RELEASED;
	stamp_t s;
	uint8_t answer = 0;

	/*
	 * If the button is pressed, move to the correct state:
	 * Releasing the button soon won't do anything, pressing longer
	 * toggles the state. As expected.
	 */
	if(IS_SET(FN_BUTTON)) {
		ENTER_CRITICAL_SECTION
			s = timer_timestamp + DISPLAY_CONF_LONGPRESS;
		LEAVE_CRITICAL_SECTION
		button_state = LONG_PRESSED;
	}

	while(1) {
		stamp_t ts;
		ENTER_CRITICAL_SECTION
			ts = timer_timestamp;
		LEAVE_CRITICAL_SECTION

		/* Force display update: */
		display_forced_process();

		switch(button_state) {
		case RELEASED:
			if(!IS_SET(FN_BUTTON)) break;
			s = ts + DISPLAY_CONF_LONGPRESS;

			_delay_ms(10);
			if(!IS_SET(FN_BUTTON)) break;

			button_state = PRESSED;
			break;

		case PRESSED:
		case LONG_PRESSED:
			if(!IS_SET(FN_BUTTON)) {
				if(button_state == PRESSED) { /* Short press: Toggle state */
					answer = answer == 0;
					display_buffer[0].lo = answer ? '1' : '0';
				} /* else:	Long press quit, nothing to do */
				button_state = BUTTON_RELEASED;
			} else if (s > ts) {	/* Press still too short for a long press */
				break;
			} else { /* Long press, leave. */
				return answer;
			}
			s = ts + DISPLAY_CONF_LONGPRESS;
		}
	}
}
